@@ -41,6 +41,12 @@
 #include <QtCore/QSharedPointer>
 #include <QtCore/QVector>
 
+#ifdef Q_OS_WIN
+#include <QtCore/qt_windows.h>
+#else
+#define IMAGE_FILE_MACHINE_ARM64 0xaa64
+#endif
+
 #include <algorithm>
 #include <iostream>
 #include <cstdio>
@@ -114,58 +120,62 @@ struct QtModuleEntry {
 };
 
 static QtModuleEntry qtModuleEntries[] = {
-    { QtBluetoothModule, "bluetooth", "Qt5Bluetooth", 0 },
+    { QtBluetoothModule, "bluetooth", "Qt5Bluetooth", nullptr },
     { QtConcurrentModule, "concurrent", "Qt5Concurrent", "qtbase" },
     { QtCoreModule, "core", "Qt5Core", "qtbase" },
     { QtDeclarativeModule, "declarative", "Qt5Declarative", "qtquick1" },
-    { QtDesignerModule, "designer", "Qt5Designer", 0 },
-    { QtDesignerComponents, "designercomponents", "Qt5DesignerComponents", 0 },
-    { QtEnginioModule, "enginio", "Enginio", 0 },
-    { QtGamePadModule, "gamepad", "Qt5Gamepad", 0 },
+    { QtDesignerModule, "designer", "Qt5Designer", nullptr },
+    { QtDesignerComponents, "designercomponents", "Qt5DesignerComponents", nullptr },
+    { QtEnginioModule, "enginio", "Enginio", nullptr },
+    { QtGamePadModule, "gamepad", "Qt5Gamepad", nullptr },
     { QtGuiModule, "gui", "Qt5Gui", "qtbase" },
     { QtHelpModule, "qthelp", "Qt5Help", "qt_help" },
     { QtMultimediaModule, "multimedia", "Qt5Multimedia", "qtmultimedia" },
     { QtMultimediaWidgetsModule, "multimediawidgets", "Qt5MultimediaWidgets", "qtmultimedia" },
     { QtMultimediaQuickModule, "multimediaquick", "Qt5MultimediaQuick_p", "qtmultimedia" },
     { QtNetworkModule, "network", "Qt5Network", "qtbase" },
-    { QtNfcModule, "nfc", "Qt5Nfc", 0 },
-    { QtOpenGLModule, "opengl", "Qt5OpenGL", 0 },
-    { QtPositioningModule, "positioning", "Qt5Positioning", 0 },
-    { QtPrintSupportModule, "printsupport", "Qt5PrintSupport", 0 },
+    { QtNfcModule, "nfc", "Qt5Nfc", nullptr },
+    { QtOpenGLModule, "opengl", "Qt5OpenGL", nullptr },
+    { QtPositioningModule, "positioning", "Qt5Positioning", nullptr },
+    { QtPrintSupportModule, "printsupport", "Qt5PrintSupport", nullptr },
     { QtQmlModule, "qml", "Qt5Qml", "qtdeclarative" },
-    { QtQmlToolingModule, "qmltooling", "qmltooling", 0 },
+    { QtQmlToolingModule, "qmltooling", "qmltooling", nullptr },
     { QtQuickModule, "quick", "Qt5Quick", "qtdeclarative" },
-    { QtQuickParticlesModule, "quickparticles", "Qt5QuickParticles", 0 },
-    { QtQuickWidgetsModule, "quickwidgets", "Qt5QuickWidgets", 0 },
+    { QtQuickParticlesModule, "quickparticles", "Qt5QuickParticles", nullptr },
+    { QtQuickWidgetsModule, "quickwidgets", "Qt5QuickWidgets", nullptr },
     { QtScriptModule, "script", "Qt5Script", "qtscript" },
     { QtScriptToolsModule, "scripttools", "Qt5ScriptTools", "qtscript" },
-    { QtSensorsModule, "sensors", "Qt5Sensors", 0 },
+    { QtSensorsModule, "sensors", "Qt5Sensors", nullptr },
     { QtSerialPortModule, "serialport", "Qt5SerialPort", "qtserialport" },
     { QtSqlModule, "sql", "Qt5Sql", "qtbase" },
-    { QtSvgModule, "svg", "Qt5Svg", 0 },
+    { QtSvgModule, "svg", "Qt5Svg", nullptr },
     { QtTestModule, "test", "Qt5Test", "qtbase" },
-    { QtWebKitModule, "webkit", "Qt5WebKit", 0 },
-    { QtWebKitWidgetsModule, "webkitwidgets", "Qt5WebKitWidgets", 0 },
-    { QtWebSocketsModule, "websockets", "Qt5WebSockets", 0 },
+    { QtWebKitModule, "webkit", "Qt5WebKit", nullptr },
+    { QtWebKitWidgetsModule, "webkitwidgets", "Qt5WebKitWidgets", nullptr },
+    { QtWebSocketsModule, "websockets", "Qt5WebSockets", nullptr },
     { QtWidgetsModule, "widgets", "Qt5Widgets", "qtbase" },
-    { QtWinExtrasModule, "winextras", "Qt5WinExtras", 0 },
+    { QtWinExtrasModule, "winextras", "Qt5WinExtras", nullptr },
     { QtXmlModule, "xml", "Qt5Xml", "qtbase" },
     { QtXmlPatternsModule, "xmlpatterns", "Qt5XmlPatterns", "qtxmlpatterns" },
-    { QtWebEngineCoreModule, "webenginecore", "Qt5WebEngineCore", 0 },
+    { QtWebEngineCoreModule, "webenginecore", "Qt5WebEngineCore", nullptr },
     { QtWebEngineModule, "webengine", "Qt5WebEngine", "qtwebengine" },
-    { QtWebEngineWidgetsModule, "webenginewidgets", "Qt5WebEngineWidgets", 0 },
-    { Qt3DCoreModule, "3dcore", "Qt53DCore", 0 },
-    { Qt3DRendererModule, "3drenderer", "Qt53DRender", 0 },
-    { Qt3DQuickModule, "3dquick", "Qt53DQuick", 0 },
-    { Qt3DQuickRendererModule, "3dquickrenderer", "Qt53DQuickRender", 0 },
-    { Qt3DInputModule, "3dinput", "Qt53DInput", 0 },
-    { Qt3DAnimationModule, "3danimation", "Qt53DAnimation", 0 },
-    { Qt3DExtrasModule, "3dextras", "Qt53DExtras", 0 },
-    { QtLocationModule, "geoservices", "Qt5Location", 0 },
-    { QtWebChannelModule, "webchannel", "Qt5WebChannel", 0 },
-    { QtTextToSpeechModule, "texttospeech", "Qt5TextToSpeech", 0 },
-    { QtSerialBusModule, "serialbus", "Qt5SerialBus", 0 },
-    { QtWebViewModule, "webview", "Qt5WebView", 0 }
+    { QtWebEngineWidgetsModule, "webenginewidgets", "Qt5WebEngineWidgets", nullptr },
+    { Qt3DCoreModule, "3dcore", "Qt53DCore", nullptr },
+    { Qt3DRendererModule, "3drenderer", "Qt53DRender", nullptr },
+    { Qt3DQuickModule, "3dquick", "Qt53DQuick", nullptr },
+    { Qt3DQuickRendererModule, "3dquickrenderer", "Qt53DQuickRender", nullptr },
+    { Qt3DInputModule, "3dinput", "Qt53DInput", nullptr },
+    { Qt3DAnimationModule, "3danimation", "Qt53DAnimation", nullptr },
+    { Qt3DExtrasModule, "3dextras", "Qt53DExtras", nullptr },
+    { QtLocationModule, "geoservices", "Qt5Location", nullptr },
+    { QtWebChannelModule, "webchannel", "Qt5WebChannel", nullptr },
+    { QtTextToSpeechModule, "texttospeech", "Qt5TextToSpeech", nullptr },
+    { QtSerialBusModule, "serialbus", "Qt5SerialBus", nullptr },
+    { QtWebViewModule, "webview", "Qt5WebView", nullptr }
+};
+
+enum QtPlugin {
+    QtVirtualKeyboardPlugin = 0x1
 };
 
 static const char webKitProcessC[] = "QtWebProcess";
@@ -180,12 +190,11 @@ static inline QString webProcessBinary(const char *binaryName, Platform p)
 static QByteArray formatQtModules(quint64 mask, bool option = false)
 {
     QByteArray result;
-    const size_t qtModulesCount = sizeof(qtModuleEntries)/sizeof(QtModuleEntry);
-    for (size_t i = 0; i < qtModulesCount; ++i) {
-        if (mask & qtModuleEntries[i].module) {
+    for (const auto &qtModule : qtModuleEntries) {
+        if (mask & qtModule.module) {
             if (!result.isEmpty())
                 result.append(' ');
-            result.append(option ? qtModuleEntries[i].option : qtModuleEntries[i].libraryName);
+            result.append(option ? qtModule.option : qtModule.libraryName);
         }
     }
     return result;
@@ -255,6 +264,7 @@ struct Options {
     bool translations = true;
     bool systemD3dCompiler = true;
     bool compilerRunTime = false;
+    unsigned disabledPlugins = 0;
     AngleDetection angleDetection = AngleDetectionAuto;
     bool softwareRasterizer = true;
     Platform platform = WindowsDesktop;
@@ -262,6 +272,7 @@ struct Options {
     quint64 disabledLibraries = 0;
     unsigned updateFileFlags = 0;
     QStringList qmlDirectories; // Project's QML files.
+    QStringList qmlImportPaths; // Custom QML module locations.
     QString directory;
     QString translationsDirectory; // Translations target directory
     QString libraryDirectory;
@@ -310,9 +321,8 @@ enum CommandLineParseFlag {
 static inline int parseArguments(const QStringList &arguments, QCommandLineParser *parser,
                                  Options *options, QString *errorMessage)
 {
-    typedef QSharedPointer<QCommandLineOption> CommandLineOptionPtr;
-    typedef QPair<CommandLineOptionPtr, quint64> OptionMaskPair;
-    typedef QVector<OptionMaskPair> OptionMaskVector;
+    using CommandLineOptionPtr = QSharedPointer<QCommandLineOption>;
+    using OptionPtrVector = QVector<CommandLineOptionPtr>;
 
     parser->setSingleDashWordOptionMode(QCommandLineParser::ParseAsLongOptions);
     parser->setApplicationDescription(QStringLiteral("Qt Deploy Tool ") + QLatin1String(QT_VERSION_STR)
@@ -378,6 +388,11 @@ static inline int parseArguments(const QStringList &arguments, QCommandLineParse
                                     QStringLiteral("directory"));
     parser->addOption(qmlDirOption);
 
+    QCommandLineOption qmlImportOption(QStringLiteral("qmlimport"),
+                                       QStringLiteral("Add the given path to the QML module search locations."),
+                                       QStringLiteral("directory"));
+    parser->addOption(qmlImportOption);
+
     QCommandLineOption noQuickImportOption(QStringLiteral("no-quick-import"),
                                            QStringLiteral("Skip deployment of Qt Quick imports."));
     parser->addOption(noQuickImportOption);
@@ -395,6 +410,10 @@ static inline int parseArguments(const QStringList &arguments, QCommandLineParse
                                              QStringLiteral("Deploy compiler runtime (Desktop only)."));
     parser->addOption(compilerRunTimeOption);
 
+    QCommandLineOption noVirtualKeyboardOption(QStringLiteral("no-virtualkeyboard"),
+                                               QStringLiteral("Disable deployment of the Virtual Keyboard."));
+    parser->addOption(noVirtualKeyboardOption);
+
     QCommandLineOption noCompilerRunTimeOption(QStringLiteral("no-compiler-runtime"),
                                              QStringLiteral("Do not deploy compiler runtime (Desktop only)."));
     parser->addOption(noCompilerRunTimeOption);
@@ -444,21 +463,22 @@ static inline int parseArguments(const QStringList &arguments, QCommandLineParse
     parser->addPositionalArgument(QStringLiteral("[files]"),
                                   QStringLiteral("Binaries or directory containing the binary."));
 
-    OptionMaskVector enabledModules;
-    OptionMaskVector disabledModules;
-    const size_t qtModulesCount = sizeof(qtModuleEntries)/sizeof(QtModuleEntry);
-    for (size_t i = 0; i < qtModulesCount; ++i) {
+    OptionPtrVector enabledModuleOptions;
+    OptionPtrVector disabledModuleOptions;
+    const int qtModulesCount = int(sizeof(qtModuleEntries) / sizeof(QtModuleEntry));
+    enabledModuleOptions.reserve(qtModulesCount);
+    disabledModuleOptions.reserve(qtModulesCount);
+    for (int i = 0; i < qtModulesCount; ++i) {
         const QString option = QLatin1String(qtModuleEntries[i].option);
         const QString name = QLatin1String(qtModuleEntries[i].libraryName);
         const QString enabledDescription = QStringLiteral("Add ") + name + QStringLiteral(" module.");
         CommandLineOptionPtr enabledOption(new QCommandLineOption(option, enabledDescription));
         parser->addOption(*enabledOption.data());
-        enabledModules.push_back(OptionMaskPair(enabledOption, qtModuleEntries[i].module));
-
+        enabledModuleOptions.append(enabledOption);
         const QString disabledDescription = QStringLiteral("Remove ") + name + QStringLiteral(" module.");
         CommandLineOptionPtr disabledOption(new QCommandLineOption(QStringLiteral("no-") + option,
                                                                    disabledDescription));
-        disabledModules.push_back(OptionMaskPair(disabledOption, qtModuleEntries[i].module));
+        disabledModuleOptions.append(disabledOption);
         parser->addOption(*disabledOption.data());
     }
 
@@ -488,6 +508,9 @@ static inline int parseArguments(const QStringList &arguments, QCommandLineParse
         return CommandLineParseError;
     }
 
+    if (parser->isSet(noVirtualKeyboardOption))
+        options->disabledPlugins |= QtVirtualKeyboardPlugin;
+
     if (parser->isSet(releaseWithDebugInfoOption))
         std::wcerr << "Warning: " << releaseWithDebugInfoOption.names().first() << " is obsolete.";
 
@@ -534,11 +557,11 @@ static inline int parseArguments(const QStringList &arguments, QCommandLineParse
 
     options->patchQt = !parser->isSet(noPatchQtOption);
 
-    for (size_t i = 0; i < qtModulesCount; ++i) {
-        if (parser->isSet(*enabledModules.at(int(i)).first.data()))
-            options->additionalLibraries |= enabledModules.at(int(i)).second;
-        if (parser->isSet(*disabledModules.at(int(i)).first.data()))
-            options->disabledLibraries |= disabledModules.at(int(i)).second;
+    for (int i = 0; i < qtModulesCount; ++i) {
+        if (parser->isSet(*enabledModuleOptions.at(i)))
+            options->additionalLibraries |= qtModuleEntries[i].module;
+        if (parser->isSet(*disabledModuleOptions.at(i)))
+            options->disabledLibraries |= qtModuleEntries[i].module;
     }
 
     // Add some dependencies
@@ -590,6 +613,9 @@ static inline int parseArguments(const QStringList &arguments, QCommandLineParse
     if (parser->isSet(qmlDirOption))
         options->qmlDirectories = parser->values(qmlDirOption);
 
+    if (parser->isSet(qmlImportOption))
+        options->qmlImportPaths = parser->values(qmlImportOption);
+
     const QString &file = posArgs.front();
     const QFileInfo fi(QDir::cleanPath(file));
     if (!fi.exists()) {
@@ -680,13 +706,14 @@ static inline bool isQtModule(const QString &libName)
 // Helper for recursively finding all dependent Qt libraries.
 static bool findDependentQtLibraries(const QString &qtBinDir, const QString &binary, Platform platform,
                                      QString *errorMessage, QStringList *result,
-                                     unsigned *wordSize = 0, bool *isDebug = 0,
-                                     int *directDependencyCount = 0, int recursionDepth = 0)
+                                     unsigned *wordSize = nullptr, bool *isDebug = nullptr,
+                                     unsigned short *machineArch = nullptr,
+                                     int *directDependencyCount = nullptr, int recursionDepth = 0)
 {
     QStringList dependentLibs;
     if (directDependencyCount)
         *directDependencyCount = 0;
-    if (!readExecutable(binary, platform, errorMessage, &dependentLibs, wordSize, isDebug)) {
+    if (!readExecutable(binary, platform, errorMessage, &dependentLibs, wordSize, isDebug, machineArch)) {
         errorMessage->prepend(QLatin1String("Unable to find dependent libraries of ") +
                               QDir::toNativeSeparators(binary) + QLatin1String(" :"));
         return false;
@@ -706,7 +733,8 @@ static bool findDependentQtLibraries(const QString &qtBinDir, const QString &bin
         *directDependencyCount = end - start;
     // Recurse
     for (int i = start; i < end; ++i)
-        if (!findDependentQtLibraries(qtBinDir, result->at(i), platform, errorMessage, result, 0, 0, 0, recursionDepth + 1))
+        if (!findDependentQtLibraries(qtBinDir, result->at(i), platform, errorMessage, result,
+                                      nullptr, nullptr, nullptr, nullptr, recursionDepth + 1))
             return false;
     return true;
 }
@@ -805,6 +833,7 @@ static const PluginModuleMapping pluginModuleMappings[] =
     {"imageformats", QtGuiModule},
     {"platforms", QtGuiModule},
     {"platforminputcontexts", QtGuiModule},
+    {"virtualkeyboard", QtGuiModule},
     {"geoservices", QtLocationModule},
     {"audio", QtMultimediaModule},
     {"mediaservice", QtMultimediaModule},
@@ -829,10 +858,9 @@ static const PluginModuleMapping pluginModuleMappings[] =
 
 static inline quint64 qtModuleForPlugin(const QString &subDirName)
 {
-    const PluginModuleMapping *end = pluginModuleMappings
-        + sizeof(pluginModuleMappings) / sizeof(pluginModuleMappings[0]);
-    const PluginModuleMapping *result =
-        std::find_if(pluginModuleMappings, end,
+    const auto end = std::end(pluginModuleMappings);
+    const auto result =
+        std::find_if(std::begin(pluginModuleMappings), end,
                      [&subDirName] (const PluginModuleMapping &m) { return subDirName == QLatin1String(m.directoryName); });
     return result != end ? result->module : 0; // "designer"
 }
@@ -851,18 +879,18 @@ static quint64 qtModule(QString module, const QString &infix)
     if (endPos > 0)
         module.truncate(endPos);
     // That should leave us with 'Qt5Core<d>'.
-    const size_t qtModulesCount = sizeof(qtModuleEntries)/sizeof(QtModuleEntry);
-    for (size_t i = 0; i < qtModulesCount; ++i) {
-        const QLatin1String libraryName(qtModuleEntries[i].libraryName);
+    for (const auto &qtModule : qtModuleEntries) {
+        const QLatin1String libraryName(qtModule.libraryName);
         if (module == libraryName
             || (module.size() == libraryName.size() + 1 && module.startsWith(libraryName))) {
-            return qtModuleEntries[i].module;
+            return qtModule.module;
         }
     }
     return 0;
 }
 
 QStringList findQtPlugins(quint64 *usedQtModules, quint64 disabledQtModules,
+                          unsigned disabledPlugins,
                           const QString &qtPluginsDirName, const QString &libraryLocation,
                           const QString &infix,
                           DebugMatchMode debugMatchModeIn, Platform platform, QString *platformPlugin)
@@ -882,6 +910,8 @@ QStringList findQtPlugins(quint64 *usedQtModules, quint64 disabledQtModules,
                 : debugMatchModeIn;
             QDir subDir(subDirFi.absoluteFilePath());
             // Filter out disabled plugins
+            if ((disabledPlugins & QtVirtualKeyboardPlugin) && subDirName == QLatin1String("virtualkeyboard"))
+                continue;
             if (disabledQtModules & QtQmlToolingModule && subDirName == QLatin1String("qmltooling"))
                 continue;
             // Filter for platform or any.
@@ -910,6 +940,11 @@ QStringList findQtPlugins(quint64 *usedQtModules, quint64 disabledQtModules,
             }
             const QStringList plugins = findSharedLibraries(subDir, platform, debugMatchMode, filter);
             for (const QString &plugin : plugins) {
+                // Filter out disabled plugins
+                if ((disabledPlugins & QtVirtualKeyboardPlugin)
+                    && plugin.startsWith(QLatin1String("qtvirtualkeyboardplugin"))) {
+                    continue;
+                }
                 const QString pluginPath = subDir.absoluteFilePath(plugin);
                 if (isPlatformPlugin)
                     *platformPlugin = pluginPath;
@@ -945,10 +980,9 @@ QStringList findQtPlugins(quint64 *usedQtModules, quint64 disabledQtModules,
 static QStringList translationNameFilters(quint64 modules, const QString &prefix)
 {
     QStringList result;
-    const size_t qtModulesCount = sizeof(qtModuleEntries)/sizeof(QtModuleEntry);
-    for (size_t i = 0; i < qtModulesCount; ++i) {
-        if ((qtModuleEntries[i].module & modules) && qtModuleEntries[i].translation) {
-            const QString name = QLatin1String(qtModuleEntries[i].translation) +
+    for (const auto &qtModule : qtModuleEntries) {
+        if ((qtModule.module & modules) && qtModule.translation) {
+            const QString name = QLatin1String(qtModule.translation) +
                                  QLatin1Char('_') +  prefix + QStringLiteral(".qm");
             if (!result.contains(name))
                 result.push_back(name);
@@ -1002,7 +1036,8 @@ static bool deployTranslations(const QString &sourcePath, quint64 usedQtModules,
             std::wcout << "Creating " << targetFile << "...\n";
         unsigned long exitCode;
         if ((options.updateFileFlags & SkipUpdateFile) == 0
-            && (!runProcess(binary, arguments, sourcePath, &exitCode, 0, 0, errorMessage) || exitCode)) {
+            && (!runProcess(binary, arguments, sourcePath, &exitCode, nullptr, nullptr, errorMessage)
+                || exitCode)) {
             return false;
         }
     } // for prefixes.
@@ -1080,7 +1115,7 @@ static QString vcRedistDir()
     return QString();
 }
 
-static QStringList compilerRunTimeLibs(Platform platform, bool isDebug, unsigned wordSize)
+static QStringList compilerRunTimeLibs(Platform platform, bool isDebug, unsigned short machineArch)
 {
     QStringList result;
     switch (platform) {
@@ -1094,24 +1129,24 @@ static QStringList compilerRunTimeLibs(Platform platform, bool isDebug, unsigned
         const QString binPath = QFileInfo(gcc).absolutePath();
         QStringList filters;
         const QString suffix = QLatin1Char('*') + sharedLibrarySuffix(platform);
-        const size_t count = sizeof(minGwRuntimes) / sizeof(minGwRuntimes[0]);
-        for (size_t i = 0; i < count; ++i)
-            filters.append(QLatin1String(minGwRuntimes[i]) + suffix);
+        for (auto minGwRuntime : minGwRuntimes)
+            filters.append(QLatin1String(minGwRuntime) + suffix);
         const QFileInfoList &dlls = QDir(binPath).entryInfoList(filters, QDir::Files);
         for (const QFileInfo &dllFi : dlls)
                 result.append(dllFi.absoluteFilePath());
     }
         break;
+#ifdef Q_OS_WIN
     case WindowsDesktop: { // MSVC/Desktop: Add redistributable packages.
         QString vcRedistDirName = vcRedistDir();
         if (vcRedistDirName.isEmpty())
              break;
         QStringList redistFiles;
         QDir vcRedistDir(vcRedistDirName);
-        const QString wordSizeString(QLatin1String(wordSize > 32 ? "x64" : "x86"));
+        const QString machineArchString = getArchString(machineArch);
         if (isDebug) {
             // Append DLLs from Debug_NonRedist\x??\Microsoft.VC<version>.DebugCRT.
-            if (vcRedistDir.cd(vcDebugRedistDir()) && vcRedistDir.cd(wordSizeString)) {
+            if (vcRedistDir.cd(vcDebugRedistDir()) && vcRedistDir.cd(machineArchString)) {
                 const QStringList names = vcRedistDir.entryList(QStringList(QStringLiteral("Microsoft.VC*.DebugCRT")), QDir::Dirs);
                 if (!names.isEmpty() && vcRedistDir.cd(names.first())) {
                     const QFileInfoList &dlls = vcRedistDir.entryInfoList(QStringList(QLatin1String("*.dll")));
@@ -1125,10 +1160,10 @@ static QStringList compilerRunTimeLibs(Platform platform, bool isDebug, unsigned
             if (!countryCodes.isEmpty()) // Pre MSVC2017
                 releaseRedistDir += QLatin1Char('/') + countryCodes.constFirst();
             QFileInfo fi(releaseRedistDir + QLatin1Char('/') + QStringLiteral("vc_redist.")
-                         + wordSizeString + QStringLiteral(".exe"));
+                         + machineArchString + QStringLiteral(".exe"));
             if (!fi.isFile()) { // Pre MSVC2017/15.5
                 fi.setFile(releaseRedistDir + QLatin1Char('/') + QStringLiteral("vcredist_")
-                           + wordSizeString + QStringLiteral(".exe"));
+                           + machineArchString + QStringLiteral(".exe"));
             }
             if (fi.isFile())
                 redistFiles.append(fi.absoluteFilePath());
@@ -1140,6 +1175,8 @@ static QStringList compilerRunTimeLibs(Platform platform, bool isDebug, unsigned
         }
         result.append(redistFiles);
     }
+        break;
+#endif // Q_OS_WIN
     default:
         break;
     }
@@ -1181,24 +1218,6 @@ static bool updateLibrary(const QString &sourceFileName, const QString &targetDi
     return true;
 }
 
-// Check for a Qt Quick Controls import path and return the version.
-// 'QtQuick/Controls' ==> 1, or 'QtQuick/Controls.2' ==> 2.
-static inline int quickControlsImportPath(const QString &ip)
-{
-    if (ip.contains(QLatin1String("Qt/labs/calendar"))
-        || ip.contains(QLatin1String("Qt/labs/folderlistmodel"))
-        || ip.contains(QLatin1String("Qt/labs/settings"))
-        || ip.contains(QLatin1String("QtQuick/Templates.2"))
-        || ip.contains(QLatin1String("QtQuick/Controls.2"))) {
-        return 2;
-    }
-    if (ip.endsWith(QLatin1String("QtQuick/Dialogs")) || ip.contains(QLatin1String("QtQuick/Dialogs/"))
-        || ip.contains(QLatin1String("QtQuick/Controls"))) {
-        return 1; // Dialogs only in v1, so far; no version number on directory: v1.
-    }
-    return 0; // Non-controls import
-}
-
 static DeployResult deploy(const Options &options,
                            const QMap<QString, QString> &qmakeVariables,
                            QString *errorMessage)
@@ -1211,7 +1230,7 @@ static DeployResult deploy(const Options &options,
     const QString libraryLocation = options.platform == Unix ? qmakeVariables.value(QStringLiteral("QT_INSTALL_LIBS")) : qtBinDir;
     const QString infix = qmakeVariables.value(QLatin1String(qmakeInfixKey));
     const int version = qtVersion(qmakeVariables);
-    Q_UNUSED(version)
+    Q_UNUSED(version);
 
     if (optVerboseLevel > 1)
         std::wcout << "Qt binaries in " << QDir::toNativeSeparators(qtBinDir) << '\n';
@@ -1219,9 +1238,10 @@ static DeployResult deploy(const Options &options,
     QStringList dependentQtLibs;
     bool detectedDebug;
     unsigned wordSize;
+    unsigned short machineArch;
     int directDependencyCount = 0;
     if (!findDependentQtLibraries(libraryLocation, options.binaries.first(), options.platform, errorMessage, &dependentQtLibs, &wordSize,
-                                  &detectedDebug, &directDependencyCount)) {
+                                  &detectedDebug, &machineArch, &directDependencyCount)) {
         return result;
     }
     for (int b = 1; b < options.binaries.size(); ++b) {
@@ -1298,6 +1318,10 @@ static DeployResult deploy(const Options &options,
     // Scan Quick2 imports
     QmlImportScanResult qmlScanResult;
     if (options.quickImports && usesQml2) {
+        // Custom list of import paths provided by user
+        QStringList qmlImportPaths = options.qmlImportPaths;
+        // Qt's own QML modules
+        qmlImportPaths << qmakeVariables.value(QStringLiteral("QT_INSTALL_QML"));
         QStringList qmlDirectories = options.qmlDirectories;
         if (qmlDirectories.isEmpty()) {
             const QString qmlDirectory = findQmlDirectory(options.platform, options.directory);
@@ -1308,7 +1332,7 @@ static DeployResult deploy(const Options &options,
             if (optVerboseLevel >= 1)
                 std::wcout << "Scanning " << QDir::toNativeSeparators(qmlDirectory) << ":\n";
             const QmlImportScanResult scanResult =
-                runQmlImportScanner(qmlDirectory, qmakeVariables.value(QStringLiteral("QT_INSTALL_QML")),
+                runQmlImportScanner(qmlDirectory, qmlImportPaths,
                                     result.directlyUsedQtLibraries & QtWidgetsModule,
                                     options.platform, debugMatchMode, errorMessage);
             if (!scanResult.ok)
@@ -1316,7 +1340,7 @@ static DeployResult deploy(const Options &options,
             qmlScanResult.append(scanResult);
             // Additional dependencies of QML plugins.
             for (const QString &plugin : qAsConst(qmlScanResult.plugins)) {
-                if (!findDependentQtLibraries(libraryLocation, plugin, options.platform, errorMessage, &dependentQtLibs, &wordSize, &detectedDebug))
+                if (!findDependentQtLibraries(libraryLocation, plugin, options.platform, errorMessage, &dependentQtLibs, &wordSize, &detectedDebug, &machineArch))
                     return result;
             }
             if (optVerboseLevel >= 1) {
@@ -1351,17 +1375,17 @@ static DeployResult deploy(const Options &options,
         findQtPlugins(&result.deployedQtLibraries,
                       // For non-QML applications, disable QML to prevent it from being pulled in by the qtaccessiblequick plugin.
                       options.disabledLibraries | (usesQml2 ? 0 : (QtQmlModule | QtQuickModule)),
+                      options.disabledPlugins,
                       qmakeVariables.value(QStringLiteral("QT_INSTALL_PLUGINS")), libraryLocation, infix,
                       debugMatchMode, options.platform, &platformPlugin);
 
     // Apply options flags and re-add library names.
     QString qtGuiLibrary;
-    const size_t qtModulesCount = sizeof(qtModuleEntries)/sizeof(QtModuleEntry);
-    for (size_t i = 0; i < qtModulesCount; ++i) {
-        if (result.deployedQtLibraries & qtModuleEntries[i].module) {
-            const QString library = libraryPath(libraryLocation, qtModuleEntries[i].libraryName, qtLibInfix, options.platform, isDebug);
+    for (const auto &qtModule : qtModuleEntries) {
+        if (result.deployedQtLibraries & qtModule.module) {
+            const QString library = libraryPath(libraryLocation, qtModule.libraryName, qtLibInfix, options.platform, isDebug);
             deployedQtLibraries.append(library);
-            if (qtModuleEntries[i].module == QtGuiModule)
+            if (qtModule.module == QtGuiModule)
                 qtGuiLibrary = library;
         }
     }
@@ -1398,7 +1422,7 @@ static DeployResult deploy(const Options &options,
         if (options.angleDetection != Options::AngleDetectionForceOff
             && (dependsOnAngle || dependsOnCombinedAngle || !dependsOnOpenGl || options.angleDetection == Options::AngleDetectionForceOn)) {
             const QString combinedAngleFullPath = qtBinDir + slash + libCombinedQtAngleName;
-            if (QFileInfo(combinedAngleFullPath).exists()) {
+            if (QFileInfo::exists(combinedAngleFullPath)) {
                 deployedQtLibraries.append(combinedAngleFullPath);
             } else {
                 const QString libGlesFullPath = qtBinDir + slash + libGlesName;
@@ -1410,7 +1434,8 @@ static DeployResult deploy(const Options &options,
                 deployedQtLibraries.append(libEglFullPath);
             }
             // Find the system D3d Compiler matching the D3D library.
-            if (options.systemD3dCompiler && !options.isWinRt()) {
+            // Any arm64 OS will be new enough to be shipped with the D3DCompiler inbox.
+            if (options.systemD3dCompiler && !options.isWinRt() && machineArch != IMAGE_FILE_MACHINE_ARM64) {
                 const QString d3dCompiler = findD3dCompiler(options.platform, qtBinDir, wordSize);
                 if (d3dCompiler.isEmpty()) {
                     std::wcerr << "Warning: Cannot find any version of the d3dcompiler DLL.\n";
@@ -1459,7 +1484,7 @@ static DeployResult deploy(const Options &options,
             options.directory : options.libraryDirectory;
         QStringList libraries = deployedQtLibraries;
         if (options.compilerRunTime)
-            libraries.append(compilerRunTimeLibs(options.platform, isDebug, wordSize));
+            libraries.append(compilerRunTimeLibs(options.platform, isDebug, machineArch));
         for (const QString &qtLib : qAsConst(libraries)) {
             if (!updateLibrary(qtLib, targetPath, options, errorMessage))
                 return result;
@@ -1591,8 +1616,8 @@ static bool deployWebEngineCore(const QMap<QString, QString> &qmakeVariables,
     const QString resourcesTargetDir(options.directory + resourcesSubDir);
     if (!createDirectory(resourcesTargetDir, errorMessage))
         return false;
-    for (size_t i = 0; i < sizeof(installDataFiles)/sizeof(installDataFiles[0]); ++i) {
-        if (!updateFile(resourcesSourceDir + QLatin1String(installDataFiles[i]),
+    for (auto installDataFile : installDataFiles) {
+        if (!updateFile(resourcesSourceDir + QLatin1String(installDataFile),
                         resourcesTargetDir, options.updateFileFlags, options.json, errorMessage)) {
             return false;
         }
@@ -1609,21 +1634,20 @@ static bool deployWebEngineCore(const QMap<QString, QString> &qmakeVariables,
         return createDirectory(options.translationsDirectory, errorMessage)
                 && updateFile(translations.absoluteFilePath(), options.translationsDirectory,
                               options.updateFileFlags, options.json, errorMessage);
-    } else {
-        // Translations have been turned off, but QtWebEngine needs at least one.
-        const QFileInfo enUSpak(translations.filePath() + QStringLiteral("/en-US.pak"));
-        if (!enUSpak.exists()) {
-            std::wcerr << "Warning: Cannot find "
-                       << QDir::toNativeSeparators(enUSpak.absoluteFilePath()) << ".\n";
-            return true;
-        }
-        const QString webEngineTranslationsDir = options.translationsDirectory + QLatin1Char('/')
-                + translations.fileName();
-        if (!createDirectory(webEngineTranslationsDir, errorMessage))
-            return false;
-        return updateFile(enUSpak.absoluteFilePath(), webEngineTranslationsDir,
-                          options.updateFileFlags, options.json, errorMessage);
     }
+    // Translations have been turned off, but QtWebEngine needs at least one.
+    const QFileInfo enUSpak(translations.filePath() + QStringLiteral("/en-US.pak"));
+    if (!enUSpak.exists()) {
+        std::wcerr << "Warning: Cannot find "
+                   << QDir::toNativeSeparators(enUSpak.absoluteFilePath()) << ".\n";
+        return true;
+    }
+    const QString webEngineTranslationsDir = options.translationsDirectory + QLatin1Char('/')
+            + translations.fileName();
+    if (!createDirectory(webEngineTranslationsDir, errorMessage))
+        return false;
+    return updateFile(enUSpak.absoluteFilePath(), webEngineTranslationsDir,
+                      options.updateFileFlags, options.json, errorMessage);
 }
 
 int main(int argc, char **argv)
@@ -1717,7 +1741,7 @@ int main(int argc, char **argv)
         else
             std::fputs(options.json->toJson().constData(), stdout);
         delete options.json;
-        options.json = 0;
+        options.json = nullptr;
     }
 
     return 0;
